import unittest
from lb_matching_tools.cleaner import MetadataCleaner


TEST_STRINGS = [
    ("recording", "Nothing fancy about this!", "Nothing fancy about this!"),
    ("recording", "Tabula Rasa (feat. Lorraine Weiss)", "Tabula Rasa"),
    ("recording", "For The Love feat. Amy True", "For The Love"),
    ("recording", "For The Love ft. Amy True", "For The Love"),
    ("recording", "For The Love ft Amy True", "For The Love"),
    ("recording", "For The Lovefeat. Amy True", "For The Lovefeat. Amy True"),
    ("recording", "For The Love feat.Amy True", "For The Love"),
    ("recording", "Amoeba (raft boy)", "Amoeba (raft boy)"),
    ("recording", "Birds Without a Feather", "Birds Without a Feather"),
    ("recording", "Don't Give up - 2001 remaster", "Don't Give up"),
    ("recording", 'Define Self-delusion at Dictionary.Com', 'Define Self-delusion at Dictionary.Com'),
    ("recording", "Running up that hill (a deal with god) - 2018 Remaster", "Running up that hill (a deal with god)"),
    ("recording", "TO STAY ALIVE [Feat. SkullyOSkully]", "TO STAY ALIVE"),
    ("recording", "Amy Winehouse, Weird Al", "Amy Winehouse, Weird Al"),
    ("recording", "!!ESIRPRUS (Ver. 2)", "!!ESIRPRUS"),
    ("recording", '"...as others see us...": I. Henry VIII (1491-1547)', '"...as others see us...": I. Henry VIII'),
    ("recording", "Herbert West – Reanimator", "Herbert West"),  # en dash
    ("recording", "Philip the Chancelor: Clavus pungens — 2vv conductus", "Philip the Chancelor: Clavus pungens"), # em dash
    ("recording", '"Beckoning Darkness" ~ Menu', '"Beckoning Darkness"'),
    ("recording", 'Big Phat at Glen Echo, MD Contradance / CALLER: Ridge Kennedy', 'Big Phat at Glen Echo, MD Contradance'),
    # Trailing punctuation (or punctuation in general) can be discarded in a consecutive step
    ("recording", 'Kikagaku Moyo/幾何学模様', 'Kikagaku Moyo/'),
    ("recording", '山地情歌 San-Di Love Song', 'San-Di Love Song'),
    ("recording", 'Madness (DJ Gollum feat. DJ Cap Remix)', 'Madness (DJ Gollum feat. DJ Cap Remix)'),
    ("recording", 'Other Place [Live]', 'Other Place'),
    ("recording", 'Other Place {Live}', 'Other Place'),
    ("recording", 'Other Place <Live>', 'Other Place'),
    ("recording", 'Other Place [boo - hiss]', 'Other Place [boo - hiss]'),
    ("recording", 'Other Place {boo - hiss}', 'Other Place {boo - hiss}'),
    ("recording", 'Other Place <boo - hiss>', 'Other Place <boo - hiss>'),
    ("recording", 'Other Place [2028 - remaster]', 'Other Place'),
    ("recording", 'Other Place {2018 - release}', 'Other Place'),
    ("recording", 'Other Place <1985 - tryout>', 'Other Place'),
    ("recording", 'T - Ball Rag', 'T - Ball Rag'),
    ("recording", ')--- ---x--- ---(', ')--- ---x--- ---('),
    ("recording", '[intentionally left blank]', '[intentionally left blank]'),
    ("recording", 'Babylumalotoroony and the Jerry Lewis Bone-A-Thons featuring Athena', 'Babylumalotoroony and the Jerry Lewis Bone-A-Thons'),
    ("recording", "Guff (Johann Sebastian - Bach)", "Guff (Johann Sebastian - Bach)"), 
    ("recording", "-10 on the Care-Meter", "-10 on the Care-Meter"), 
    ("recording", "264 - Das Herz", "264 - Das Herz"),
    # This should not be split, because a / ~ - inside () [] {} <> should not be split
    ("recording", "(You're So Square) Baby I Don't Care (Movie Edit, 2013, Take 16/2021, Take 6) Binaural", 
                  "(You're So Square) Baby I Don't Care (Movie Edit, 2013, Take 16/2021, Take 6) Binaural"),
    ("recording", "this random text (quite worthy!)", "this random text (quite worthy!)"),
    ("recording", "this random text (released 2018)", "this random text"),
    ("recording", "this random text (demo session)", "this random text"),
    ("recording", "this random text (tryout tape)", "this random text"),
    ("recording", "I Miss You (Howie B mix)", "I Miss You (Howie B mix)"),
    ("recording", "Run Jozi (Godly)", "Run Jozi (Godly)"),
    ("recording", "Όσο Και Να Σ' Αγαπάω (Υπ' Ευθύνη Μου)", "Όσο Και Να Σ' Αγαπάω (Υπ' Ευθύνη Μου)"),
    ("recording", "Όσο Και Να Σ' Αγαπάω (2018 remaster)", "Όσο Και Να Σ' Αγαπάω"),
    ("recording", "Όσο Και Να Σ' Αγαπάω - Remix 2023", "Όσο Και Να Σ' Αγαπάω"),

    ("artist", "Amy Winehouse, Weird Al", "Amy Winehouse"),
    ("artist", "((( Punchline Paradise )))", "((( Punchline Paradise )))"),
    ("artist", 'Dan No Ura 壇ノ浦の戦い', 'Dan No Ura'),
    ("artist", '4 Dan No Ura 壇ノ浦の戦い', '4 Dan No Ura'),
    ("artist", "Babylumalotoroony and the Jerry Lewis Bone-A-Thons featuring Athena", "Babylumalotoroony and the Jerry Lewis Bone-A-Thons"),
    ("artist", "Wilhelm Schneider-Clauss & Heimersdorfer Kinderchor", "Wilhelm Schneider-Clauss"),
    ("artist", "Jim Foster with The Kountry Kut-Ups", "Jim Foster")
]

def suite():
    suite = unittest.TestSuite()
    suite.addTests(TestMetadataCleaner(clean_type, input, output) for clean_type, input, output in TEST_STRINGS)
    return suite


class TestMetadataCleaner(unittest.TestCase):
    def __init__(self, clean_type, input, output):
        super(TestMetadataCleaner, self).__init__()
        self.input = input
        self.output = output
        self.clean_type = clean_type

    def runTest(self):
        cl = MetadataCleaner()
        if self.clean_type == "recording":
            self.assertEqual(cl.clean_recording(self.input), self.output)
        else:
            self.assertEqual(cl.clean_artist(self.input), self.output)



import re
import regex

from lb_matching_tools.guff import GUFF_PAREN_WORDS

SYMBOLS = "1234567890!@#$%^&*()-=_+[]{};\"|;'\\<>?/.,~`"


def hyphen_split_check(original, cleaned):
    if len(cleaned) < len(original) / 3:
        return False
    return True


class MetadataCleaner:

    RECORDING_EXPRESSIONS = [
        # This track is crap (2018 remaster)
        r"(?P<title>.+?)\s+(?P<enclosed>\(.+\)|\[.+\]|\{.+\}|\<.+\>)$",

        # Tabula Rasa (feat. Lorraine Weiss)
        # TO STAY ALIVE [Feat. SkullyOSkully]
        # For The Love feat. Amy True
        # For The Love ft. Amy True
        # For The Love ft Amy True
        # Birds Without a Feather -> Nothing!
        r"(?P<title>.+?)\s+?(?P<feat>[\[\(]?(?:feat(?:uring)?|ft)\b\.?)\s*?(?P<artists>.+)\s*",

        # Don't Give up - 2001 remaster
        r"(?P<title>.+?)(?:\s+?[\u2010\u2012\u2013\u2014~/-])(?![^(]*\))(?P<dash>.*)",
    ]

    ARTIST_EXPRESSIONS = [
        # Amy Winehouse, Weird Al Yankovic
        r"(?P<title>.+?)(?:\s*?,)(?P<comma>.*)",

        # Wilhelm Schneider Clauss with Heimersdorfer Kinderchor
        # Wilhelm Schneider Clauss & Heimersdorfer Kinderchor
        r"(?P<title>.+?)(?:\s+?(&|with))(?P<dash>.*)",
    ]

    def __init__(self, preferred_script="Latin"):
        self.recording_expressions = [re.compile(exp, re.IGNORECASE) for exp in self.RECORDING_EXPRESSIONS]
        self.artist_expressions = [re.compile(exp, re.IGNORECASE) for exp in self.ARTIST_EXPRESSIONS]
        self.foreign_script_expression = regex.compile(r"[^\p{Script=Common}\p{Script=" + preferred_script + r"}]+")
        self.paren_guff_expression = re.compile(r"(20[0-9]{2}|19[0-9]{2})")
        self.letter_expression = regex.compile(r"\p{Letter}")

    def drop_foreign_chars(self, text: str):
        remaining_text = regex.sub(self.foreign_script_expression, "", text).strip()
        # only return the remaining text if it still contains at least one letter
        if regex.search(self.letter_expression, remaining_text):
            return remaining_text
        else:
            return text

    def is_paren_text_likely_guff(self, paren_text):
        """
            Given some text found in a paren, guess to see if its guff (useless meta info) or relevant info for matching a track.
            Return True if the given text is considered guff.
        """

        paren_text = paren_text.lower()
        before_len = len(paren_text)

        # Remove guff words from text
        for guff in GUFF_PAREN_WORDS:
            paren_text = paren_text.replace(guff, "")

        # Check to see if we find a year (19## or 20##). If we do, remove it
        paren_text = self.paren_guff_expression.sub("", paren_text)
        replaced = before_len - len(paren_text)

        # Calculate the number of "characters" vs "symbols" for what is leftover
        chars = 0
        guff_chars = replaced
        for ch in paren_text.lower():

            if ch in SYMBOLS:
                guff_chars += 1

            if regex.match(self.letter_expression, ch):
                chars += 1

        if guff_chars <= chars:
            return False

        return True

    def paren_checker(self, text):
        """
            Check to see if the cleaned fragment contains a split paren (), {}, [], <>. If it does, return False, otherwise True.
        """

        if text.count("(") != text.count(")"):
            return False

        if text.count("<") != text.count(">"):
            return False

        if text.count("[") != text.count("]"):
            return False

        if text.count("{") != text.count("}"):
            return False

        return True

    def clean_recording(self, text: str):
        """
            Run the metadata cleaner against a recording name. Returns the cleaned string, which may be unchanged from the given string.
        """
        cleaned = ""
        for i, exp in enumerate(self.recording_expressions):
            m = exp.match(text)
            if m is not None:
                is_guff_text = False
                try:
                    paren_text = m.group('enclosed')
                    is_guff_text = self.is_paren_text_likely_guff(paren_text[1:-1])
                    if is_guff_text:
                        return m.groups()[0]
                    else:
                        continue
                except IndexError:
                    cleaned = m.groups()[0]

                # Check to make sure the regexes didn't produce crap
                # This is ugly.
                # Could be avoided by using a list of keywords which indicate crap and e.g. baking them into a lookahead expression.
                if i == 2:
                    if not hyphen_split_check(text, cleaned):
                        cleaned = text

                if not self.paren_checker(cleaned):
                    cleaned = text

        if not len(cleaned):
            cleaned = text

        return self.drop_foreign_chars(cleaned)

    def clean_artist(self, text: str):
        """
            Run the metadata cleaner against an artist name. Returns the cleaned string, which may be unchanged from the given string.
        """

        cleaned = self.clean_recording(text)
        if text != cleaned:
            return cleaned

        for i, exp in enumerate(self.artist_expressions):
            m = exp.match(text)
            if m is not None:
                return m.groups()[0]

        return text



GUFF_PAREN_WORDS = [
    "a cappella",
    "acoustic",
    "bonus",  
    "censored",
    "clean",
    "club",                                                                     
    "clubmix",
    "composition",
    "cut",
    "dance",
    "demo",
    "dialogue",
    "dirty",                                  
    "edit",
    "excerpt",
    "explicit",
    "extended",
    "instrumental",
    "interlude",
    "intro",
    "karaoke",
    "live",
    "long",                   
    "main",
    "maxi",
    "megamix",
    "mix",
    "mono",
    "official",
    "orchestral",
    "original",
    "outro",
    "outtake",
    "outtakes",
    "piano",
    "quadraphonic",
    "radio",
    "rap",
    "re-edit",
    "reedit",
    "refix",
    "rehearsal",
    "reinterpreted",
    "released",
    "release",
    "remake",
    "remastered",
    "remaster",
    "master",  
    "remix",
    "remixed",
    "remode",
    "reprise",                                 
    "rework",
    "reworked",
    "rmx",
    "session",
    "short",
    "single",
    "skit",
    "stereo",
    "studio",
    "take",
    "takes",
    "tape",
    "track",
    "tryout",
    "uncensored",
    "unknown",
    "unplugged",
    "untitled",
    "version",
    "ver",
    "video",
    "vocal",
    "vs",                             
    "with",
    "without"
]
